import concurrent.futures as cf
import copy
import sys
import threading

import pytest

import sentry_sdk
from sentry_sdk import start_span, start_transaction
from sentry_sdk.feature_flags import FlagBuffer, add_feature_flag
from tests.conftest import ApproxDict


def test_featureflags_integration(sentry_init, capture_events, uninstall_integration):
    sentry_init()

    add_feature_flag("hello", False)
    add_feature_flag("world", True)
    add_feature_flag("other", False)

    events = capture_events()
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 1
    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
            {"flag": "world", "result": True},
            {"flag": "other", "result": False},
        ]
    }


@pytest.mark.asyncio
async def test_featureflags_integration_spans_async(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
    )
    events = capture_events()

    add_feature_flag("hello", False)

    try:
        with sentry_sdk.start_span(name="test-span"):
            with sentry_sdk.start_span(name="test-span-2"):
                raise ValueError("something wrong!")
    except ValueError as e:
        sentry_sdk.capture_exception(e)

    found = False
    for event in events:
        if "exception" in event.keys():
            assert event["contexts"]["flags"] == {
                "values": [
                    {"flag": "hello", "result": False},
                ]
            }
            found = True

    assert found, "No event with exception found"


def test_featureflags_integration_spans_sync(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=1.0,
    )
    events = capture_events()

    add_feature_flag("hello", False)

    try:
        with sentry_sdk.start_span(name="test-span"):
            with sentry_sdk.start_span(name="test-span-2"):
                raise ValueError("something wrong!")
    except ValueError as e:
        sentry_sdk.capture_exception(e)

    found = False
    for event in events:
        if "exception" in event.keys():
            assert event["contexts"]["flags"] == {
                "values": [
                    {"flag": "hello", "result": False},
                ]
            }
            found = True

    assert found, "No event with exception found"


def test_featureflags_integration_threaded(
    sentry_init, capture_events, uninstall_integration
):
    sentry_init()
    events = capture_events()

    # Capture an eval before we split isolation scopes.
    add_feature_flag("hello", False)

    def task(flag_key):
        # Creates a new isolation scope for the thread.
        # This means the evaluations in each task are captured separately.
        with sentry_sdk.isolation_scope():
            add_feature_flag(flag_key, False)
            # use a tag to identify to identify events later on
            sentry_sdk.set_tag("task_id", flag_key)
            sentry_sdk.capture_exception(Exception("something wrong!"))

    # Run tasks in separate threads
    with cf.ThreadPoolExecutor(max_workers=2) as pool:
        pool.map(task, ["world", "other"])

    # Capture error in original scope
    sentry_sdk.set_tag("task_id", "0")
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 3
    events.sort(key=lambda e: e["tags"]["task_id"])

    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
        ]
    }
    assert events[1]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
            {"flag": "other", "result": False},
        ]
    }
    assert events[2]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
            {"flag": "world", "result": False},
        ]
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python3.7 or higher")
def test_featureflags_integration_asyncio(
    sentry_init, capture_events, uninstall_integration
):
    asyncio = pytest.importorskip("asyncio")

    sentry_init()
    events = capture_events()

    # Capture an eval before we split isolation scopes.
    add_feature_flag("hello", False)

    async def task(flag_key):
        # Creates a new isolation scope for the thread.
        # This means the evaluations in each task are captured separately.
        with sentry_sdk.isolation_scope():
            add_feature_flag(flag_key, False)
            # use a tag to identify to identify events later on
            sentry_sdk.set_tag("task_id", flag_key)
            sentry_sdk.capture_exception(Exception("something wrong!"))

    async def runner():
        return asyncio.gather(task("world"), task("other"))

    asyncio.run(runner())

    # Capture error in original scope
    sentry_sdk.set_tag("task_id", "0")
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 3
    events.sort(key=lambda e: e["tags"]["task_id"])

    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
        ]
    }
    assert events[1]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
            {"flag": "other", "result": False},
        ]
    }
    assert events[2]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": False},
            {"flag": "world", "result": False},
        ]
    }


def test_flag_tracking():
    """Assert the ring buffer works."""
    buffer = FlagBuffer(capacity=3)
    buffer.set("a", True)
    flags = buffer.get()
    assert len(flags) == 1
    assert flags == [{"flag": "a", "result": True}]

    buffer.set("b", True)
    flags = buffer.get()
    assert len(flags) == 2
    assert flags == [{"flag": "a", "result": True}, {"flag": "b", "result": True}]

    buffer.set("c", True)
    flags = buffer.get()
    assert len(flags) == 3
    assert flags == [
        {"flag": "a", "result": True},
        {"flag": "b", "result": True},
        {"flag": "c", "result": True},
    ]

    buffer.set("d", False)
    flags = buffer.get()
    assert len(flags) == 3
    assert flags == [
        {"flag": "b", "result": True},
        {"flag": "c", "result": True},
        {"flag": "d", "result": False},
    ]

    buffer.set("e", False)
    buffer.set("f", False)
    flags = buffer.get()
    assert len(flags) == 3
    assert flags == [
        {"flag": "d", "result": False},
        {"flag": "e", "result": False},
        {"flag": "f", "result": False},
    ]

    # Test updates
    buffer.set("e", True)
    buffer.set("e", False)
    buffer.set("e", True)
    flags = buffer.get()
    assert flags == [
        {"flag": "d", "result": False},
        {"flag": "f", "result": False},
        {"flag": "e", "result": True},
    ]

    buffer.set("d", True)
    flags = buffer.get()
    assert flags == [
        {"flag": "f", "result": False},
        {"flag": "e", "result": True},
        {"flag": "d", "result": True},
    ]


def test_flag_buffer_concurrent_access():
    buffer = FlagBuffer(capacity=100)
    error_occurred = False

    def writer():
        for i in range(1_000_000):
            buffer.set(f"key_{i}", True)

    def reader():
        nonlocal error_occurred

        try:
            for _ in range(1000):
                copy.deepcopy(buffer)
        except RuntimeError:
            error_occurred = True

    writer_thread = threading.Thread(target=writer)
    reader_thread = threading.Thread(target=reader)

    writer_thread.start()
    reader_thread.start()

    writer_thread.join(timeout=5)
    reader_thread.join(timeout=5)

    # This should always be false. If this ever fails we know we have concurrent access to a
    # shared resource. When deepcopying we should have exclusive access to the underlying
    # memory.
    assert error_occurred is False


def test_flag_limit(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)

    events = capture_events()

    with start_transaction(name="hi"):
        with start_span(op="foo", name="bar"):
            add_feature_flag("0", True)
            add_feature_flag("1", True)
            add_feature_flag("2", True)
            add_feature_flag("3", True)
            add_feature_flag("4", True)
            add_feature_flag("5", True)
            add_feature_flag("6", True)
            add_feature_flag("7", True)
            add_feature_flag("8", True)
            add_feature_flag("9", True)
            add_feature_flag("10", True)

    (event,) = events
    assert event["spans"][0]["data"] == ApproxDict(
        {
            "flag.evaluation.0": True,
            "flag.evaluation.1": True,
            "flag.evaluation.2": True,
            "flag.evaluation.3": True,
            "flag.evaluation.4": True,
            "flag.evaluation.5": True,
            "flag.evaluation.6": True,
            "flag.evaluation.7": True,
            "flag.evaluation.8": True,
            "flag.evaluation.9": True,
        }
    )
    assert "flag.evaluation.10" not in event["spans"][0]["data"]



import json

import pytest
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application, HTTPError, RequestHandler

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.integrations.tornado import TornadoIntegration


@pytest.fixture
def tornado_testcase(request):
    # Take the unittest class provided by tornado and manually call its setUp
    # and tearDown.
    #
    # The pytest plugins for tornado seem too complicated to use, as they for
    # some reason assume I want to write my tests in async code.
    def inner(app):
        class TestBogus(AsyncHTTPTestCase):
            def get_app(self):
                return app

            def bogustest(self):
                # We need to pass a valid test method name to the ctor, so this
                # is the method. It does nothing.
                pass

        self = TestBogus("bogustest")
        self.setUp()
        request.addfinalizer(self.tearDown)
        return self

    return inner


class CrashingHandler(RequestHandler):
    def get(self):
        sentry_sdk.get_isolation_scope().set_tag("foo", "42")
        1 / 0

    def post(self):
        sentry_sdk.get_isolation_scope().set_tag("foo", "43")
        1 / 0


class CrashingWithMessageHandler(RequestHandler):
    def get(self):
        capture_message("hi")
        1 / 0


class HelloHandler(RequestHandler):
    async def get(self):
        sentry_sdk.get_isolation_scope().set_tag("foo", "42")

        return b"hello"

    async def post(self):
        sentry_sdk.get_isolation_scope().set_tag("foo", "43")

        return b"hello"


class ChildSpanHandler(RequestHandler):
    def get(self):
        with sentry_sdk.traces.start_span(name="child-span"):
            pass
        self.write("ok")


def test_basic(tornado_testcase, sentry_init, capture_events):
    sentry_init(integrations=[TornadoIntegration()], send_default_pii=True)
    events = capture_events()
    client = tornado_testcase(Application([(r"/hi", CrashingHandler)]))

    response = client.fetch(
        "/hi?foo=bar", headers={"Cookie": "name=value; name2=value2; name3=value3"}
    )
    assert response.code == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    assert exception["mechanism"]["type"] == "tornado"

    request = event["request"]
    host = request["headers"]["Host"]
    assert event["request"] == {
        "env": {"REMOTE_ADDR": "127.0.0.1"},
        "headers": {
            "Accept-Encoding": "gzip",
            "Connection": "close",
            "Cookie": "name=value; name2=value2; name3=value3",
            **request["headers"],
        },
        "cookies": {"name": "value", "name2": "value2", "name3": "value3"},
        "method": "GET",
        "query_string": "foo=bar",
        "url": "http://{host}/hi".format(host=host),
    }

    assert event["tags"] == {"foo": "42"}
    assert (
        event["transaction"]
        == "tests.integrations.tornado.test_tornado.CrashingHandler.get"
    )
    assert event["transaction_info"] == {"source": "component"}

    assert not sentry_sdk.get_isolation_scope()._tags


@pytest.mark.parametrize("send_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize(
    "handler,code",
    [
        (CrashingHandler, 500),
        (HelloHandler, 200),
    ],
)
def test_transactions(
    tornado_testcase,
    sentry_init,
    capture_events,
    capture_items,
    handler,
    code,
    span_streaming,
    send_pii,
):
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("event", "span")
    else:
        events = capture_events()

    client = tornado_testcase(Application([(r"/hi", handler)]))

    if span_streaming:
        with sentry_sdk.traces.start_span(name="client") as span:
            request_headers = dict(span._iter_headers())
    else:
        with start_transaction(name="client") as span:
            pass
        request_headers = dict(span.iter_headers())

    response = client.fetch(
        "/hi?foo=bar", method="POST", body=b"heyoo", headers=request_headers
    )
    assert response.code == code

    sentry_sdk.flush()

    if span_streaming:
        spans = [i.payload for i in items if i.type == "span"]
        errors = [i.payload for i in items if i.type == "event"]

        client_segment, server_segment = spans

        if code == 500:
            assert len(errors) == 1
            server_error = errors[0]
            assert server_error["exception"]["values"][0]["type"] == "ZeroDivisionError"
            assert (
                server_error["transaction"]
                == "tests.integrations.tornado.test_tornado.CrashingHandler.post"
            )
            assert server_error["transaction_info"] == {"source": "component"}
            assert (
                server_error["contexts"]["trace"]["trace_id"]
                == server_segment["trace_id"]
            )

        expected_handler = (
            "tests.integrations.tornado.test_tornado.HelloHandler.post"
            if code == 200
            else "tests.integrations.tornado.test_tornado.CrashingHandler.post"
        )
        assert server_segment["name"] == expected_handler
        assert server_segment["attributes"]["sentry.segment.name.source"] == "component"
        assert server_segment["attributes"]["http.request.method"] == "POST"
        assert server_segment["attributes"]["http.request.body.data"] == "heyoo"
        assert server_segment["attributes"]["http.response.status_code"] == code
        assert server_segment["status"] == ("ok" if code == 200 else "error")
        assert client_segment["trace_id"] == server_segment["trace_id"]

        if send_pii:
            assert server_segment["attributes"]["url.query"] == "foo=bar"
            assert server_segment["attributes"]["url.full"].endswith("/hi?foo=bar")
            assert server_segment["attributes"]["url.full"].startswith("http://")
            assert server_segment["attributes"]["url.path"] == "/hi"
        else:
            assert "url.query" not in server_segment["attributes"]
            assert "url.full" not in server_segment["attributes"]
            assert "url.path" not in server_segment["attributes"]
    else:
        if code == 200:
            client_tx, server_tx = events
            server_error = None
        else:
            client_tx, server_error, server_tx = events

        assert client_tx["type"] == "transaction"
        assert client_tx["transaction"] == "client"
        assert client_tx["transaction_info"] == {
            "source": "custom"
        }  # because this is just the start_transaction() above.

        if server_error is not None:
            assert server_error["exception"]["values"][0]["type"] == "ZeroDivisionError"
            assert (
                server_error["transaction"]
                == "tests.integrations.tornado.test_tornado.CrashingHandler.post"
            )
            assert server_error["transaction_info"] == {"source": "component"}

        if code == 200:
            assert (
                server_tx["transaction"]
                == "tests.integrations.tornado.test_tornado.HelloHandler.post"
            )
        else:
            assert (
                server_tx["transaction"]
                == "tests.integrations.tornado.test_tornado.CrashingHandler.post"
            )

        assert server_tx["transaction_info"] == {"source": "component"}
        assert server_tx["type"] == "transaction"

        request = server_tx["request"]
        host = request["headers"]["Host"]
        expected_request = {
            "env": {"REMOTE_ADDR": "127.0.0.1"},
            "headers": {
                "Accept-Encoding": "gzip",
                "Connection": "close",
                **request["headers"],
            },
            "method": "POST",
            "query_string": "foo=bar",
            "data": {"heyoo": [""]},
            "url": "http://{host}/hi".format(host=host),
        }
        if send_pii:
            expected_request["cookies"] = {}
        assert server_tx["request"] == expected_request

        assert (
            client_tx["contexts"]["trace"]["trace_id"]
            == server_tx["contexts"]["trace"]["trace_id"]
        )

        if server_error is not None:
            assert (
                server_error["contexts"]["trace"]["trace_id"]
                == server_tx["contexts"]["trace"]["trace_id"]
            )


def test_400_not_logged(tornado_testcase, sentry_init, capture_events):
    sentry_init(integrations=[TornadoIntegration()])
    events = capture_events()

    class CrashingHandler(RequestHandler):
        def get(self):
            raise HTTPError(400, "Oops")

    client = tornado_testcase(Application([(r"/", CrashingHandler)]))

    response = client.fetch("/")
    assert response.code == 400

    assert not events


def test_user_auth(tornado_testcase, sentry_init, capture_events):
    sentry_init(integrations=[TornadoIntegration()], send_default_pii=True)
    events = capture_events()

    class UserHandler(RequestHandler):
        def get(self):
            1 / 0

        def get_current_user(self):
            return 42

    class NoUserHandler(RequestHandler):
        def get(self):
            1 / 0

    client = tornado_testcase(
        Application([(r"/auth", UserHandler), (r"/noauth", NoUserHandler)])
    )

    # has user
    response = client.fetch("/auth")
    assert response.code == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"

    assert event["user"] == {"is_authenticated": True}

    events.clear()

    # has no user
    response = client.fetch("/noauth")
    assert response.code == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"

    assert "user" not in event


def test_formdata(tornado_testcase, sentry_init, capture_events):
    sentry_init(integrations=[TornadoIntegration()], send_default_pii=True)
    events = capture_events()

    class FormdataHandler(RequestHandler):
        def post(self):
            raise ValueError(json.dumps(sorted(self.request.body_arguments)))

    client = tornado_testcase(Application([(r"/form", FormdataHandler)]))

    response = client.fetch(
        "/form?queryarg=1",
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        body=b"field1=value1&field2=value2",
    )

    assert response.code == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["value"] == '["field1", "field2"]'
    assert event["request"]["data"] == {"field1": ["value1"], "field2": ["value2"]}


def test_json(tornado_testcase, sentry_init, capture_events):
    sentry_init(integrations=[TornadoIntegration()], send_default_pii=True)
    events = capture_events()

    class FormdataHandler(RequestHandler):
        def post(self):
            raise ValueError(json.dumps(sorted(self.request.body_arguments)))

    client = tornado_testcase(Application([(r"/form", FormdataHandler)]))

    response = client.fetch(
        "/form?queryarg=1",
        method="POST",
        headers={"Content-Type": "application/json"},
        body=b"""
        {"foo": {"bar": 42}}
        """,
    )

    assert response.code == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["value"] == "[]"
    assert event
    assert event["request"]["data"] == {"foo": {"bar": 42}}


def test_error_has_new_trace_context_performance_enabled(
    tornado_testcase, sentry_init, capture_events
):
    """
    Check if an 'trace' context is added to errros and transactions when performance monitoring is enabled.
    """
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    client = tornado_testcase(Application([(r"/hi", CrashingWithMessageHandler)]))
    client.fetch("/hi")

    (msg_event, error_event, transaction_event) = events

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert "trace" in transaction_event["contexts"]
    assert "trace_id" in transaction_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
        == transaction_event["contexts"]["trace"]["trace_id"]
    )


def test_error_has_new_trace_context_performance_disabled(
    tornado_testcase, sentry_init, capture_events
):
    """
    Check if an 'trace' context is added to errros and transactions when performance monitoring is disabled.
    """
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=None,  # this is the default, just added for clarity
    )
    events = capture_events()

    client = tornado_testcase(Application([(r"/hi", CrashingWithMessageHandler)]))
    client.fetch("/hi")

    (msg_event, error_event) = events

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
    )


def test_error_has_existing_trace_context_performance_enabled(
    tornado_testcase, sentry_init, capture_events
):
    """
    Check if an 'trace' context is added to errros and transactions
    from the incoming 'sentry-trace' header when performance monitoring is enabled.
    """
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    headers = {"sentry-trace": sentry_trace_header}

    client = tornado_testcase(Application([(r"/hi", CrashingWithMessageHandler)]))
    client.fetch("/hi", headers=headers)

    (msg_event, error_event, transaction_event) = events

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert "trace" in transaction_event["contexts"]
    assert "trace_id" in transaction_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
        == transaction_event["contexts"]["trace"]["trace_id"]
        == "471a43a4192642f0b136d5159a501701"
    )


def test_error_has_existing_trace_context_performance_disabled(
    tornado_testcase, sentry_init, capture_events
):
    """
    Check if an 'trace' context is added to errros and transactions
    from the incoming 'sentry-trace' header when performance monitoring is disabled.
    """
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=None,  # this is the default, just added for clarity
    )
    events = capture_events()

    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    headers = {"sentry-trace": sentry_trace_header}

    client = tornado_testcase(Application([(r"/hi", CrashingWithMessageHandler)]))
    client.fetch("/hi", headers=headers)

    (msg_event, error_event) = events

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
        == "471a43a4192642f0b136d5159a501701"
    )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    tornado_testcase, sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = tornado_testcase(Application([(r"/hi", CrashingHandler)]))

    client.fetch(
        "/hi?foo=bar", headers={"Cookie": "name=value; name2=value2; name3=value3"}
    )

    sentry_sdk.flush()

    if span_streaming:
        (segment,) = [i.payload for i in items]
        assert segment["attributes"]["sentry.origin"] == "auto.http.tornado"
    else:
        (_, event) = events
        assert event["contexts"]["trace"]["origin"] == "auto.http.tornado"


@pytest.mark.parametrize("send_default_pii", [True, False])
def test_user_ip_address_on_all_spans(
    tornado_testcase, sentry_init, capture_items, send_default_pii
):
    sentry_init(
        integrations=[TornadoIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    client = tornado_testcase(Application([(r"/hi", ChildSpanHandler)]))
    client.fetch("/hi")

    sentry_sdk.flush()

    child_span, server_span = [item.payload for item in items]

    if send_default_pii:
        assert server_span["attributes"]["user.ip_address"] == "127.0.0.1"
        assert child_span["attributes"]["user.ip_address"] == "127.0.0.1"
    else:
        assert "user.ip_address" not in server_span["attributes"]
        assert "user.ip_address" not in child_span["attributes"]



"""
Tests need pytest-asyncio installed.

Tests need a local MySQL instance running. This can be done using:
```sh
docker run --rm --name some-mysql -e MYSQL_ROOT_PASSWORD=root -e MYSQL_DATABASE=test -d -p 3306:3306 mysql:8.0
```

The tests use the following credentials to establish a database connection.
"""

import datetime
import os
import warnings
from contextlib import contextmanager
from unittest import mock

import aiomysql
import pytest
import pytest_asyncio

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.aiomysql import AioMySQLIntegration
from sentry_sdk.tracing_utils import record_sql_queries
from tests.conftest import ApproxDict

MYSQL_HOST = os.getenv("SENTRY_PYTHON_TEST_MYSQL_HOST", "localhost")
MYSQL_PORT = int(os.getenv("SENTRY_PYTHON_TEST_MYSQL_PORT", "3306"))
MYSQL_USER = os.getenv("SENTRY_PYTHON_TEST_MYSQL_USER", "root")
MYSQL_PASSWORD = os.getenv("SENTRY_PYTHON_TEST_MYSQL_PASSWORD", "root")
MYSQL_DB_BASE = os.getenv("SENTRY_PYTHON_TEST_MYSQL_DB", "test")


def _get_db_name():
    pid = os.getpid()
    return f"{MYSQL_DB_BASE}_{pid}"


MYSQL_DB = _get_db_name()

CRUMBS_CONNECT = {
    "category": "query",
    "data": ApproxDict(
        {
            "db.name": MYSQL_DB,
            "db.system": "mysql",
            "db.user": MYSQL_USER,
            "server.address": MYSQL_HOST,
            "server.port": MYSQL_PORT,
        }
    ),
    "message": "connect",
    "type": "default",
}


@pytest_asyncio.fixture(autouse=True)
async def _clean_mysql():
    conn = await aiomysql.connect(
        host=MYSQL_HOST,
        port=MYSQL_PORT,
        user=MYSQL_USER,
        password=MYSQL_PASSWORD,
        autocommit=True,
    )
    try:
        async with conn.cursor() as cur:
            # Suppress MySQL warnings about unknown tables / existing databases
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", Warning)
                await cur.execute(f"CREATE DATABASE IF NOT EXISTS `{MYSQL_DB}`")
                await cur.execute(f"USE `{MYSQL_DB}`")
                await cur.execute("DROP TABLE IF EXISTS users")
            await cur.execute(
                """
                CREATE TABLE users(
                    id INT AUTO_INCREMENT PRIMARY KEY,
                    name VARCHAR(255),
                    password VARCHAR(255),
                    dob DATE
                )
                """
            )
    finally:
        conn.close()


def _connect_args():
    return {
        "host": MYSQL_HOST,
        "port": MYSQL_PORT,
        "user": MYSQL_USER,
        "password": MYSQL_PASSWORD,
        "db": MYSQL_DB,
        "autocommit": True,
    }


@pytest.mark.asyncio
async def test_connect(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())
    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [CRUMBS_CONNECT]


@pytest.mark.asyncio
async def test_execute(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())

    async with conn.cursor() as cur:
        await cur.execute(
            "INSERT INTO users(name, password, dob) VALUES ('Alice', 'pw', '1990-12-25')",
        )
        await cur.execute(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
        )
        await cur.execute("SELECT * FROM users WHERE name = %s", ("Bob",))
        row = await cur.fetchone()
        assert row[1] == "Bob"

    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {},
            "message": "INSERT INTO users(name, password, dob) VALUES ('Alice', 'pw', '1990-12-25')",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT * FROM users WHERE name = %s",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
async def test_execute_many(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())

    async with conn.cursor() as cur:
        await cur.executemany(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            [
                ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
                ("Alice", "pw", datetime.date(1990, 12, 25)),
            ],
        )

    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {"db.executemany": True},
            "message": "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
async def test_execute_many_non_insert(sentry_init, capture_events) -> None:
    """Test executemany with non-INSERT queries (falls back to row-by-row)."""
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())

    async with conn.cursor() as cur:
        # Pre-populate users table
        await cur.execute(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            ("Alice", "pw1", datetime.date(1990, 1, 1)),
        )
        await cur.execute(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            ("Bob", "pw2", datetime.date(1991, 2, 2)),
        )
        # Non-INSERT executemany — uses row-by-row fallback internally
        await cur.executemany(
            "UPDATE users SET password = %s WHERE name = %s",
            [
                ("new_pw_1", "Alice"),
                ("new_pw_2", "Bob"),
            ],
        )

    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    # Should have: connect + INSERT*2 + single executemany span (no double-recording)
    crumbs = event["breadcrumbs"]["values"]
    query_crumbs = [c for c in crumbs if c["category"] == "query"]
    executemany_crumbs = [
        c for c in query_crumbs if c.get("data", {}).get("db.executemany")
    ]
    # Only ONE executemany breadcrumb — no duplicates from internal execute calls
    assert len(executemany_crumbs) == 1
    assert "UPDATE users SET password = %s" in executemany_crumbs[0]["message"]


@pytest.mark.asyncio
async def test_record_params(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration(record_params=True)],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())

    async with conn.cursor() as cur:
        await cur.execute(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
        )

    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {
                "db.params": ["Bob", "secret_pw", "datetime.date(1984, 3, 1)"],
                "db.paramstyle": "format",
            },
            "message": "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
async def test_cursor_context_manager(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())

    async with conn.cursor() as cur:
        await cur.executemany(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            [
                ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
                ("Alice", "pw", datetime.date(1990, 12, 25)),
            ],
        )
        await cur.execute(
            "SELECT * FROM users WHERE dob > %s",
            (datetime.date(1970, 1, 1),),
        )
        rows = await cur.fetchall()
        assert len(rows) == 2

    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    crumbs = event["breadcrumbs"]["values"]
    assert crumbs[0] == CRUMBS_CONNECT
    assert crumbs[1]["category"] == "query"
    assert "INSERT" in crumbs[1]["message"]
    assert crumbs[2]["category"] == "query"
    assert "SELECT" in crumbs[2]["message"]


@pytest.mark.asyncio
async def test_cursor_async_iteration(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn = await aiomysql.connect(**_connect_args())

    async with conn.cursor() as cur:
        await cur.execute(
            "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
            ("Charlie", "pw3", datetime.date(1995, 5, 15)),
        )
        await cur.execute("SELECT * FROM users WHERE name = %s", ("Charlie",))
        async for row in cur:
            assert row[1] == "Charlie"

    conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    crumbs = event["breadcrumbs"]["values"]
    assert crumbs[0] == CRUMBS_CONNECT
    assert len([c for c in crumbs if c["category"] == "query"]) >= 2


@pytest.mark.asyncio
async def test_connection_pool(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AioMySQLIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    pool_size = 2

    pool = await aiomysql.create_pool(
        host=MYSQL_HOST,
        port=MYSQL_PORT,
        user=MYSQL_USER,
        password=MYSQL_PASSWORD,
        db=MYSQL_DB,
        autocommit=True,
        minsize=pool_size,
        maxsize=pool_size,
    )

    async with pool.acquire() as conn:
        async with conn.cursor() as cur:
            await cur.execute(
                "INSERT INTO users(name, password, dob) VALUES (%s, %s, %s)",
                ("Dave", "pw4", datetime.date(1988, 7, 20)),
            )
            await cur.execute("SELECT * FROM users WHERE name = %s", ("Dave",))
            row = await cur.fetchone()
            assert row is not None
            assert row[1] == "Dave"

    pool.close()
    await pool.wait_closed()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    crumbs = event["breadcrumbs"]["values"]
    # Verify queries were captured
    query_crumbs = [c for c in crumbs if c["category"] == "query"]
    assert len(query_crumbs) >= 2  # INSERT + SELECT

    # Verify connect spans were created for pool connections
    connect_crumbs = [c for c in crumbs if c.get("message") == "connect"]
    assert len(connect_crumbs) >= pool_size  # One connect span per pooled connection
    for crumb in connect_crumbs:
        assert crumb["data"]["db.system"] == "mysql"
        assert crumb["data"]["server.address"] == MYSQL_HOST


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_disabled(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_options = {
        "integrations": [AioMySQLIntegration()],
        "traces_sample_rate": 1.0,
        "enable_db_query_source": False,
        "db_query_source_threshold_ms": 0,
        "_experiments": {
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    }

    sentry_init(**sentry_options)

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["is_segment"] is True
        assert connect_span["name"] == "connect"
        assert insert_span["name"].startswith("INSERT INTO")

        data = insert_span.get("attributes", {})

        assert "code.line.number" not in data
        assert "code.namespace" not in data
        assert "code.file.path" not in data
        assert "code.function" not in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILEPATH not in data
        assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.asyncio
@pytest.mark.parametrize("enable_db_query_source", [None, True])
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_enabled(
    sentry_init, capture_events, capture_items, enable_db_query_source, span_streaming
):
    sentry_options = {
        "integrations": [AioMySQLIntegration()],
        "traces_sample_rate": 1.0,
        "db_query_source_threshold_ms": 0,
        "_experiments": {
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    }
    if enable_db_query_source is not None:
        sentry_options["enable_db_query_source"] = enable_db_query_source

    sentry_init(**sentry_options)

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["is_segment"] is True
        assert connect_span["name"] == "connect"
        assert insert_span["name"].startswith("INSERT INTO")

        data = insert_span.get("attributes", {})

        assert "code.line.number" in data
        assert "code.namespace" in data
        assert "code.file.path" in data
        assert "code.function" in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["is_segment"] is True
        assert connect_span["name"] == "connect"
        assert insert_span["name"].startswith("INSERT INTO")

        data = insert_span.get("attributes", {})

        assert "code.line.number" in data
        assert "code.namespace" in data
        assert "code.file.path" in data
        assert "code.function" in data

        assert type(data.get("code.line.number")) == int
        assert data.get("code.line.number") > 0
        assert data.get("code.namespace") == "tests.integrations.aiomysql.test_aiomysql"
        assert data.get("code.file.path").endswith(
            "tests/integrations/aiomysql/test_aiomysql.py"
        )

        is_relative_path = data.get("code.file.path")[0] != os.sep
        assert is_relative_path

        assert data.get("code.function") == "test_query_source"
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert (
            data.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.aiomysql.test_aiomysql"
        )
        assert data.get(SPANDATA.CODE_FILEPATH).endswith(
            "tests/integrations/aiomysql/test_aiomysql.py"
        )

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
        assert is_relative_path

        assert data.get(SPANDATA.CODE_FUNCTION) == "test_query_source"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_no_query_source_if_duration_too_short(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=100,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        @contextmanager
        def fake_record_sql_queries_streaming(*args, **kwargs):
            with record_sql_queries(*args, **kwargs) as span:
                pass
            span._start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span._end_timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
            yield span

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                with mock.patch(
                    "sentry_sdk.integrations.aiomysql.record_sql_queries",
                    fake_record_sql_queries_streaming,
                ):
                    await cur.execute(
                        "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                    )

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"
        data = insert_span.get("attributes", {})

        assert "code.line.number" not in data
        assert "code.namespace" not in data
        assert "code.file.path" not in data
        assert "code.function" not in data
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn = await aiomysql.connect(**_connect_args())

            @contextmanager
            def fake_record_sql_queries(*args, **kwargs):
                with record_sql_queries(*args, **kwargs) as span:
                    pass
                span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                span.timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
                yield span

            async with conn.cursor() as cur:
                with mock.patch(
                    "sentry_sdk.integrations.aiomysql.record_sql_queries",
                    fake_record_sql_queries,
                ):
                    await cur.execute(
                        "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                    )

            conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO not in data
        assert SPANDATA.CODE_NAMESPACE not in data
        assert SPANDATA.CODE_FILEPATH not in data
        assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_if_duration_over_threshold(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=100,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        @contextmanager
        def fake_record_sql_queries_streaming(*args, **kwargs):
            with record_sql_queries(*args, **kwargs) as span:
                span._start_timestamp = datetime.datetime(
                    2024, 1, 1, microsecond=0, tzinfo=datetime.timezone.utc
                )
                yield span

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                with mock.patch(
                    "sentry_sdk.integrations.aiomysql.record_sql_queries",
                    fake_record_sql_queries_streaming,
                ):
                    await cur.execute(
                        "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                    )

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"

        data = insert_span.get("attributes", {})

        assert "code.line.number" in data
        assert "code.namespace" in data
        assert "code.file.path" in data
        assert "code.function" in data

        assert type(data.get("code.line.number")) == int
        assert data.get("code.line.number") > 0
        assert data.get("code.namespace") == "tests.integrations.aiomysql.test_aiomysql"
        assert data.get("code.file.path").endswith(
            "tests/integrations/aiomysql/test_aiomysql.py"
        )

        is_relative_path = data.get("code.file.path")[0] != os.sep
        assert is_relative_path

        assert (
            data.get("code.function") == "test_query_source_if_duration_over_threshold"
        )
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn = await aiomysql.connect(**_connect_args())

            @contextmanager
            def fake_record_sql_queries(*args, **kwargs):
                with record_sql_queries(*args, **kwargs) as span:
                    pass
                span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                span.timestamp = datetime.datetime(2024, 1, 1, microsecond=100001)
                yield span

            async with conn.cursor() as cur:
                with mock.patch(
                    "sentry_sdk.integrations.aiomysql.record_sql_queries",
                    fake_record_sql_queries,
                ):
                    await cur.execute(
                        "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                    )

            conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")

        data = span.get("data", {})

        assert SPANDATA.CODE_LINENO in data
        assert SPANDATA.CODE_NAMESPACE in data
        assert SPANDATA.CODE_FILEPATH in data
        assert SPANDATA.CODE_FUNCTION in data

        assert type(data.get(SPANDATA.CODE_LINENO)) == int
        assert data.get(SPANDATA.CODE_LINENO) > 0
        assert (
            data.get(SPANDATA.CODE_NAMESPACE)
            == "tests.integrations.aiomysql.test_aiomysql"
        )
        assert data.get(SPANDATA.CODE_FILEPATH).endswith(
            "tests/integrations/aiomysql/test_aiomysql.py"
        )

        is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
        assert is_relative_path

        assert (
            data.get(SPANDATA.CODE_FUNCTION)
            == "test_query_source_if_duration_over_threshold"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_span_origin(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute("SELECT 1")
                await cur.execute("SELECT 2")

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        non_segment_spans = [s for s in spans if not s.get("is_segment")]
        segment_spans = [s for s in spans if s.get("is_segment")]

        assert len(segment_spans) == 1
        assert segment_spans[0]["attributes"]["sentry.origin"] == "manual"

        for span in non_segment_spans:
            assert span["attributes"]["sentry.origin"] == "auto.db.aiomysql"
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute("SELECT 1")
                await cur.execute("SELECT 2")

            conn.close()

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"

        for span in event["spans"]:
            assert span["origin"] == "auto.db.aiomysql"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_multiline_query_description_normalized(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    """
                    SELECT
                        id,
                        name
                    FROM
                        users
                    WHERE
                        name = 'Alice'
                    """
                )

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        select_spans = [
            s
            for s in spans
            if not s.get("is_segment") and "SELECT" in s.get("name", "")
        ]
        assert len(select_spans) == 1
        assert (
            select_spans[0]["name"] == "SELECT id, name FROM users WHERE name = 'Alice'"
        )
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute(
                    """
                    SELECT
                        id,
                        name
                    FROM
                        users
                    WHERE
                        name = 'Alice'
                    """
                )

            conn.close()

        (event,) = events

        spans = [
            s
            for s in event["spans"]
            if s["op"] == "db" and "SELECT" in s.get("description", "")
        ]
        assert len(spans) == 1
        assert (
            spans[0]["description"] == "SELECT id, name FROM users WHERE name = 'Alice'"
        )


@pytest.mark.asyncio
async def test_before_send_transaction_sees_normalized_description(
    sentry_init, capture_events
):
    def before_send_transaction(event, hint):
        for span in event.get("spans", []):
            desc = span.get("description", "")
            if "SELECT id, name FROM users" in desc:
                span["description"] = "filtered"
        return event

    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        before_send_transaction=before_send_transaction,
    )
    events = capture_events()

    with start_transaction(name="test_transaction"):
        conn = await aiomysql.connect(**_connect_args())

        async with conn.cursor() as cur:
            await cur.execute(
                """
                SELECT
                    id,
                    name
                FROM
                    users
                """
            )

        conn.close()

    (event,) = events
    spans = [
        s
        for s in event["spans"]
        if s["op"] == "db" and "filtered" in s.get("description", "")
    ]

    assert len(spans) == 1
    assert spans[0]["description"] == "filtered"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_db_data_on_spans(
    sentry_init, capture_events, capture_items, span_streaming
):
    """Test that database connection data is properly set on spans."""
    sentry_init(
        integrations=[AioMySQLIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute("SELECT 1")

            conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        query_span = [
            s
            for s in spans
            if not s.get("is_segment") and "SELECT" in s.get("name", "")
        ][0]
        data = query_span.get("attributes", {})
        assert data.get(SPANDATA.DB_SYSTEM_NAME) == "mysql"
        assert data.get(SPANDATA.SERVER_ADDRESS) == MYSQL_HOST
        assert data.get(SPANDATA.SERVER_PORT) == MYSQL_PORT
        assert data.get(SPANDATA.DB_NAMESPACE) == MYSQL_DB
        assert data.get(SPANDATA.DB_USER) == MYSQL_USER
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn = await aiomysql.connect(**_connect_args())

            async with conn.cursor() as cur:
                await cur.execute("SELECT 1")

            conn.close()

        (event,) = events

        db_spans = [s for s in event["spans"] if s["op"] == "db"]
        assert len(db_spans) > 0

        query_span = [s for s in db_spans if "SELECT" in s.get("description", "")][0]
        assert query_span["data"].get(SPANDATA.DB_SYSTEM) == "mysql"
        assert query_span["data"].get(SPANDATA.SERVER_ADDRESS) == MYSQL_HOST
        assert query_span["data"].get(SPANDATA.SERVER_PORT) == MYSQL_PORT
        assert query_span["data"].get(SPANDATA.DB_NAME) == MYSQL_DB
        assert query_span["data"].get(SPANDATA.DB_USER) == MYSQL_USER



import os
import sys

import pytest

from sentry_sdk.ai.utils import _normalize_data
from sentry_sdk.consts import EndpointType
from sentry_sdk.utils import (
    AnnotatedValue,
    BadDsn,
    Dsn,
    exceptions_from_error_tuple,
    filename_for_module,
    from_base64,
    iter_event_stacktraces,
    safe_repr,
    set_in_app_in_frames,
    strip_string,
    to_base64,
)

try:
    import hypothesis.strategies as st
    from hypothesis import given
except ImportError:
    pass
else:
    any_string = st.one_of(st.binary(), st.text())

    @given(x=any_string)
    def test_safe_repr_never_broken_for_strings(x):
        r = safe_repr(x)
        assert isinstance(r, str)
        assert "broken repr" not in r


def test_safe_repr_regressions():
    assert "лошадь" in safe_repr("лошадь")


@pytest.mark.parametrize("prefix", ("", "abcd", "лошадь"))
@pytest.mark.parametrize("character", "\x00\x07\x1b\n")
def test_safe_repr_non_printable(prefix, character):
    """Check that non-printable characters are escaped"""
    string = prefix + character
    assert character not in safe_repr(string)
    assert character not in safe_repr(string.encode("utf-8"))


def test_abs_path():
    """Check if abs_path is actually an absolute path. This can happen either
    with eval/exec like here, or when the file in the frame is relative to
    __main__"""

    code = compile("1/0", "test.py", "exec")
    try:
        exec(code, {})
    except Exception:
        exceptions = exceptions_from_error_tuple(sys.exc_info())

    (exception,) = exceptions
    frame1, frame2 = frames = exception["stacktrace"]["frames"]

    for frame in frames:
        assert os.path.abspath(frame["abs_path"]) == frame["abs_path"]

    assert frame1["filename"] == "tests/utils/test_general.py"
    assert frame2["filename"] == "test.py"


def test_filename():
    x = filename_for_module

    assert x("bogus", "bogus") == "bogus"

    assert x("bogus", "bogus.pyc") == "bogus.py"

    assert x("os", os.__file__) == "os.py"

    assert x("foo.bar", "path/to/foo/bar.py") == "path/to/foo/bar.py"

    import sentry_sdk.utils

    assert x("sentry_sdk.utils", sentry_sdk.utils.__file__) == "sentry_sdk/utils.py"


def test_filename_module_file_is_none():
    class DummyModule:
        __file__ = None

    os.sys.modules["foo"] = DummyModule()

    assert filename_for_module("foo.bar", "path/to/foo/bar.py") == "path/to/foo/bar.py"


@pytest.mark.parametrize(
    "given,expected_envelope",
    [
        (
            "https://foobar@sentry.io/123",
            "https://sentry.io/api/123/envelope/",
        ),
        (
            "https://foobar@sentry.io/bam/123",
            "https://sentry.io/bam/api/123/envelope/",
        ),
        (
            "https://foobar@sentry.io/bam/baz/123",
            "https://sentry.io/bam/baz/api/123/envelope/",
        ),
    ],
)
def test_parse_dsn_paths(given, expected_envelope):
    dsn = Dsn(given)
    auth = dsn.to_auth()
    assert auth.get_api_url() == expected_envelope
    assert auth.get_api_url(EndpointType.ENVELOPE) == expected_envelope


@pytest.mark.parametrize(
    "given,expected",
    [
        ("https://foobar@sentry.io/123", None),
        ("https://foobar@o1234.ingest.sentry.io/123", "1234"),
    ],
)
def test_parse_dsn_org_id(given, expected):
    dsn = Dsn(given)
    assert dsn.org_id == expected


@pytest.mark.parametrize(
    "dsn",
    [
        "https://foobar@sentry.io"
        "https://foobar@sentry.io/"
        "https://foobar@sentry.io/asdf"
        "https://foobar@sentry.io/asdf/"
        "https://foobar@sentry.io/asdf/123/"
    ],
)
def test_parse_invalid_dsn(dsn):
    with pytest.raises(BadDsn):
        dsn = Dsn(dsn)


@pytest.mark.parametrize(
    "dsn,error_message",
    [
        ("foo://barbaz@sentry.io", "Unsupported scheme 'foo'"),
        ("https://foobar@", "Missing hostname"),
        ("https://@sentry.io", "Missing public key"),
    ],
)
def test_dsn_validations(dsn, error_message):
    with pytest.raises(BadDsn) as e:
        dsn = Dsn(dsn)

    assert str(e.value) == error_message


@pytest.mark.parametrize(
    "frame,in_app_include,in_app_exclude,project_root,resulting_frame",
    [
        [
            {
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
            },
            None,
            None,
            None,
            {
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
            },
            None,
            None,
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": True,
            },
            None,
            None,
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": True,
            },
        ],
        [
            {
                "abs_path": "C:\\Users\\winuser\\AppData\\Roaming\\Python\\Python35\\site-packages\\fastapi\\routing.py",
            },
            None,
            None,
            None,
            {
                "abs_path": "C:\\Users\\winuser\\AppData\\Roaming\\Python\\Python35\\site-packages\\fastapi\\routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/usr/lib/python2.7/dist-packages/fastapi/routing.py",
            },
            None,
            None,
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/usr/lib/python2.7/dist-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            None,
            None,
            None,
            {
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
        ],
        [
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            None,
            None,
            None,
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
        ],
        # include
        [
            {
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
            },
            ["fastapi"],
            None,
            None,
            {
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,  # because there is no module set
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
            },
            ["fastapi"],
            None,
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": True,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,
            },
            ["fastapi"],
            None,
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "abs_path": "C:\\Users\\winuser\\AppData\\Roaming\\Python\\Python35\\site-packages\\fastapi\\routing.py",
            },
            ["fastapi"],
            None,
            None,
            {
                "abs_path": "C:\\Users\\winuser\\AppData\\Roaming\\Python\\Python35\\site-packages\\fastapi\\routing.py",
                "in_app": False,  # because there is no module set
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/usr/lib/python2.7/dist-packages/fastapi/routing.py",
            },
            ["fastapi"],
            None,
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/usr/lib/python2.7/dist-packages/fastapi/routing.py",
                "in_app": True,
            },
        ],
        [
            {
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            ["fastapi"],
            None,
            None,
            {
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
        ],
        [
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            ["fastapi"],
            None,
            None,
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
        ],
        # exclude
        [
            {
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
            },
            None,
            ["main"],
            None,
            {
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
            },
            None,
            ["main"],
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": True,
            },
            None,
            ["main"],
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/home/ubuntu/fastapi/.venv/lib/python3.10/site-packages/fastapi/routing.py",
                "in_app": True,
            },
        ],
        [
            {
                "abs_path": "C:\\Users\\winuser\\AppData\\Roaming\\Python\\Python35\\site-packages\\fastapi\\routing.py",
            },
            None,
            ["main"],
            None,
            {
                "abs_path": "C:\\Users\\winuser\\AppData\\Roaming\\Python\\Python35\\site-packages\\fastapi\\routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
                "abs_path": "/usr/lib/python2.7/dist-packages/fastapi/routing.py",
            },
            None,
            ["main"],
            None,
            {
                "module": "fastapi.routing",
                "abs_path": "/usr/lib/python2.7/dist-packages/fastapi/routing.py",
                "in_app": False,
            },
        ],
        [
            {
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            None,
            ["main"],
            None,
            {
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
        ],
        [
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            None,
            ["main"],
            None,
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
                "in_app": False,
            },
        ],
        [
            {
                "module": "fastapi.routing",
            },
            None,
            None,
            None,
            {
                "module": "fastapi.routing",
            },
        ],
        [
            {
                "module": "fastapi.routing",
            },
            ["fastapi"],
            None,
            None,
            {
                "module": "fastapi.routing",
                "in_app": True,
            },
        ],
        [
            {
                "module": "fastapi.routing",
            },
            None,
            ["fastapi"],
            None,
            {
                "module": "fastapi.routing",
                "in_app": False,
            },
        ],
        # with project_root set
        [
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            None,
            None,
            "/home/ubuntu/fastapi",
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
                "in_app": True,
            },
        ],
        [
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            ["main"],
            None,
            "/home/ubuntu/fastapi",
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
                "in_app": True,
            },
        ],
        [
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
            },
            None,
            ["main"],
            "/home/ubuntu/fastapi",
            {
                "module": "main",
                "abs_path": "/home/ubuntu/fastapi/main.py",
                "in_app": False,
            },
        ],
    ],
)
def test_set_in_app_in_frames(
    frame, in_app_include, in_app_exclude, project_root, resulting_frame
):
    new_frames = set_in_app_in_frames(
        [frame],
        in_app_include=in_app_include,
        in_app_exclude=in_app_exclude,
        project_root=project_root,
    )

    assert new_frames[0] == resulting_frame


def test_iter_stacktraces():
    assert set(
        iter_event_stacktraces(
            {
                "threads": {"values": [{"stacktrace": 1}]},
                "stacktrace": 2,
                "exception": {"values": [{"stacktrace": 3}]},
            }
        )
    ) == {1, 2, 3}


@pytest.mark.parametrize(
    ("original", "base64_encoded"),
    [
        # ascii only
        ("Dogs are great!", "RG9ncyBhcmUgZ3JlYXQh"),
        # emoji
        ("🐶", "8J+Qtg=="),
        # non-ascii
        (
            "Καλό κορίτσι, Μάιζεϊ!",
            "zprOsc67z4wgzrrOv8+Bzq/PhM+DzrksIM6czqzOuc62zrXPiiE=",
        ),
        # mix of ascii and non-ascii
        (
            "Of margir hundar! Ég geri ráð fyrir að ég þurfi stærra rúm.",
            "T2YgbWFyZ2lyIGh1bmRhciEgw4lnIGdlcmkgcsOhw7AgZnlyaXIgYcOwIMOpZyDDvnVyZmkgc3TDpnJyYSByw7ptLg==",
        ),
    ],
)
def test_successful_base64_conversion(original, base64_encoded):
    # all unicode characters should be handled correctly
    assert to_base64(original) == base64_encoded
    assert from_base64(base64_encoded) == original

    # "to" and "from" should be inverses
    assert from_base64(to_base64(original)) == original
    assert to_base64(from_base64(base64_encoded)) == base64_encoded


@pytest.mark.parametrize(
    "input",
    [
        1231,  # incorrect type
        True,  # incorrect type
        [],  # incorrect type
        {},  # incorrect type
        None,  # incorrect type
        "yayfordogs",  # wrong length
        "#dog",  # invalid ascii character
        "🐶",  # non-ascii character
    ],
)
def test_failed_base64_conversion(input):
    # conversion from base64 should fail if given input of the wrong type or
    # input which isn't a valid base64 string
    assert from_base64(input) is None

    # any string can be converted to base64, so only type errors will cause
    # failures
    if not isinstance(input, str):
        assert to_base64(input) is None


@pytest.mark.parametrize(
    "input,max_length,result",
    [
        [None, None, None],
        ["a" * 256, None, "a" * 256],
        [
            "a" * 257,
            256,
            AnnotatedValue(
                value="a" * 253 + "...",
                metadata={"len": 257, "rem": [["!limit", "x", 253, 256]]},
            ),
        ],
        ["éééé", None, "éééé"],
        [
            "éééé",
            5,
            AnnotatedValue(
                value="é...", metadata={"len": 8, "rem": [["!limit", "x", 2, 5]]}
            ),
        ],
        [
            "\udfff\udfff\udfff\udfff\udfff\udfff",
            5,
            AnnotatedValue(
                value="\udfff\udfff...",
                metadata={
                    "len": 6,
                    "rem": [["!limit", "x", 5 - 3, 5]],
                },
            ),
        ],
    ],
)
def test_strip_string(input, max_length, result):
    assert strip_string(input, max_length) == result


def test_normalize_data_with_pydantic_class():
    """Test that _normalize_data handles Pydantic model classes"""

    class TestClass:
        name: str = None

        def __init__(self, name: str):
            self.name = name

        def model_dump(self):
            return {"name": self.name}

    # Test with class (should NOT call model_dump())
    result = _normalize_data(TestClass)
    assert result == "<ClassType: TestClass>"

    # Test with instance (should call model_dump())
    instance = TestClass(name="test")
    result = _normalize_data(instance)
    assert result == {"name": "test"}

    # Test with dict containing class
    result = _normalize_data({"schema": TestClass, "count": 5})
    assert result == {"schema": "<ClassType: TestClass>", "count": 5}
